import unittest
from cwcwidth import wcwidth, wcswidth


class Tests(unittest.TestCase):
    def exec_test(
        self, phrase, individual_expected_lengths, expected_length=None, n=None
    ):
        if expected_length is None:
            expected_length = sum(individual_expected_lengths)
        self.assertEqual(
            individual_expected_lengths,
            tuple(map(wcwidth, phrase if n is None else phrase[:n])),
        )
        self.assertEqual(expected_length, wcswidth(phrase, n))

    def test_hello_jp(self):
        """Width of Japanese phrase: コンニチハ, セカイ!"""
        self.exec_test("コンニチハ, セカイ!", (2, 2, 2, 2, 2, 1, 1, 2, 2, 2, 1))

    def test_wcswidth_substr(self):
        """Test wcswidth() optional 2nd parameter, ``n``."""
        self.exec_test("コンニチハ, セカイ!", (2, 2, 2, 2, 2, 1, 1), n=7)

    def test_null_width_0(self):
        """NULL (0) reports width 0."""
        self.exec_test("abc\x00def", (1, 1, 1, 0, 1, 1, 1))
        self.exec_test("abc\x00def", (1, 1, 1, 0, 1, 1, 1), n=7)

    def test_control_c0_width_negative_1(self):
        """CSI (Control sequence initiate) reports width -1."""
        self.exec_test("\x1b[0m", (-1, 1, 1, 1), expected_length=-1)

    def test_combining_width_negative_1(self):
        """Simple test combining reports total width of 4."""
        self.exec_test("--\u05bf--", (1, 1, 0, 1, 1))

    def test_combining_cafe(self):
        """Phrase cafe + COMBINING ACUTE ACCENT is café of length 4."""
        self.exec_test("cafe\u0301", (1, 1, 1, 1, 0))

    def test_combining_enclosing(self):
        """CYRILLIC CAPITAL LETTER A + COMBINING CYRILLIC HUNDRED THOUSANDS SIGN is А҈ of length 1."""
        expect_length_each = (1, 0)
        self.exec_test("\u0410\u0488", (1, 0))

    def test_combining_spacing(self):
        """Balinese kapal (ship) is ᬓᬨᬮ᭄ of length 4."""
        self.exec_test("\u1B13\u1B28\u1B2E\u1B44", (1, 1, 1, 1))



# Copyright 2021 Sebastian Ramacher <sebastian@ramacher.at>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from libc.stddef cimport wchar_t, size_t
from cpython.mem cimport PyMem_Free

cdef extern from "Python.h":
    wchar_t* PyUnicode_AsWideCharString(object, Py_ssize_t*) except NULL
    Py_ssize_t PyUnicode_AsWideChar(object, wchar_t*, Py_ssize_t)

cdef extern from "wcwidth_compat.h" nogil:
    int wcswidth(const wchar_t*, size_t)
    int wcwidth(wchar_t)
    size_t wcslen(const wchar_t*)


cdef int wcswidth_loop(const wchar_t* s, size_t n):
    cdef size_t idx
    cdef int v
    cdef ret = 0
    for idx in range(n):
        v = wcwidth(s[idx])
        if v == -1:
            return -1
        ret += v
    return ret


def c_wcswidth(str pwcs not None, n=None):
    """Return the printable length of a unicode character on a terminal.

    Note that this function slightly deviates from wcswidth(3) behavior when the string includes
    null characters. As strings are not null terminated, they are treated as characters of width 0
    and processing continues until the end of the string.

    See wcswidth(3) for more details.
    """

    cdef Py_ssize_t actual_length
    cdef wchar_t* s = PyUnicode_AsWideCharString(pwcs, &actual_length)
    cdef size_t cn = actual_length
    cdef int ret = 0

    if n is not None and n < cn:
        cn = <size_t>n

    if <size_t>actual_length != wcslen(s):
        # In this case pwcs contains a null character. libc's wcwidth (and other string processing
        # functions) will stop when encountering a null character, but in Python the null character
        # will just be skipped. So in this case we will emulate wcwidth's behavior and sum up all
        # the width of all characters individually
        ret = wcswidth_loop(s, cn)
    else:
        ret = wcswidth(s, cn)
    PyMem_Free(s)
    return ret


def c_wcwidth(str wc not None):
    """Return the printable length of a unicode character on a terminal.

    See wcwidth(3) for more details.
    """

    if len(wc) != 1:
        raise ValueError("Expected one unicode character")

    cdef wchar_t c
    PyUnicode_AsWideChar(wc, &c, 1)
    return wcwidth(c)



#!/usr/bin/python3

import platform
import os.path
from setuptools import setup, Extension

try:
    from Cython.Build import cythonize

    have_cython = True
except ImportError:
    have_cython = False


def read(name):
    with open(os.path.join(os.path.dirname(__file__), name), encoding="utf-8") as f:
        return f.read()


extension_sources = ["cwcwidth/_impl.pyx" if have_cython else "cwcwidth/_impl.c"]
if platform.system() in ("Windows", "Darwin"):
    extension_sources.append("cwcwidth/wcwidth.c")
    define_macros = [
        ("USE_MK_WCWIDTH", None),
    ]
else:
    define_macros = [
        ("_XOPEN_SOURCE", "600"),
    ]

ext_modules = [
    Extension(
        "cwcwidth._impl",
        extension_sources,
        define_macros=define_macros,
    )
]

if have_cython:
    ext_modules = cythonize(ext_modules, language_level=3)
    setup_requires = ["Cython >= 0.28"]
else:
    setup_requires = []


setup(
    name="cwcwidth",
    version="0.1.1",
    description="Python bindings for wc(s)width",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    author="Sebastian Ramacher",
    author_email="sebastian@ramacher.at",
    url="https://github.com/sebastinas/cwcwidth",
    license="Expat",
    ext_modules=ext_modules,
    packages=["cwcwidth"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">= 3.6",
    setup_requires=setup_requires,
)
